import flask
import json
import logging
import time;

LOG = logging.getLogger('api')
CONTROLLER = None
app = flask.Flask('concord232')
LOG.info("API Code Loaded")

def show_zone(zone):
    return {
        'partition': zone['partition_number'],
        'area': zone['area_number'],
        'group': zone['group_number'],
        'number': zone['zone_number'],
        'name': zone['zone_text'],
        'state': zone['zone_state'],
        'type': zone['zone_type'],
        #'bypassed': zone.bypassed,
        #'condition_flags': zone.condition_flags,
        #'type_flags': zone.type_flags,
    }


def show_partition(partition):
    return {
        'number': partition['partition_number'],
        'area': partition['area_number'],
        'arming_level': partition['arming_level'],
        'arming_level_code': partition['arming_level_code'],
        'partition_text': partition['partition_text'],
        'zones': sum(z['partition_number'] == partition['partition_number'] for z in CONTROLLER.zones.values()),


    }

@app.route('/panel')
def index_panel():
    try:
        result = json.dumps({
            'panel': CONTROLLER.panel
            })
        return flask.Response(result,
                              mimetype='application/json')
    except Exception as e:
        LOG.exception('Failed to index zones')

@app.route('/zones')
def index_zones():
    try:
        if not bool(CONTROLLER.zones):
            CONTROLLER.request_zones();
            
        while not bool(CONTROLLER.zones):
            time.sleep(0.25)

        result = json.dumps({
            'zones': [show_zone(zone) for zone in CONTROLLER.zones.values()]})
        return flask.Response(result,
                              mimetype='application/json')
    except Exception as e:
        LOG.exception('Failed to index zones')




@app.route('/partitions')
def index_partitions():
    try:
        if not bool(CONTROLLER.partitions):
            CONTROLLER.request_partitions();
         
        while not bool(CONTROLLER.partitions):
            time.sleep(0.25)

        result = json.dumps({
            'partitions': [show_partition(partition)
                           for partition in CONTROLLER.partitions.values()]})
        return flask.Response(result,
                              mimetype='application/json')
    except Exception as e:
        LOG.exception('Failed to index partitions')


@app.route('/command')
def command():
    args = flask.request.args
    if args.get('cmd') == 'arm':
        option = args.get('option')
        if args.get('level') == 'stay':
            CONTROLLER.arm_stay(option)
        elif args.get('level') == 'away':
            CONTROLLER.arm_away(option)
    elif args.get('cmd') == 'disarm':
        CONTROLLER.disarm(args.get('master_pin'))
    elif args.get('cmd') == 'keys':
        CONTROLLER.send_keys(args.get('keys'),args.get('group'))
    return flask.Response()

@app.route('/version')
def get_version():
    return flask.Response(json.dumps({'version': '1.1'}),
                          mimetype='application/json')

@app.route('/equipment')
def get_equipment():
    CONTROLLER.request_all_equipment()
    return flask.Response()    


@app.route('/all_data')
def get_all_data():
    CONTROLLER.request_dynamic_data_refresh()
    return flask.Response()    



try:
    import ConfigParser as configparser
except ImportError:
    import configparser
import email
import email.mime
import email.mime.text
import email.utils
import smtplib


class MissingEmailConfig(Exception):
    pass


def _send_system_email(config, subject, recips, body):
    try:
        fromaddr = config.get('email', 'fromaddr')
        smtphost = config.get('email', 'smtphost')
    except (configparser.NoOptionError,
            configparser.NoSectionError):
        raise MissingEmailConfig()

    msg = email.mime.text.MIMEText(body)
    msg['Subject'] = subject
    msg['From'] = fromaddr
    msg['Date'] = email.utils.formatdate()
    msg['Message-Id'] = email.utils.make_msgid('concord232')
    for addr in recips:
        msg['To'] = addr

    smtp = smtplib.SMTP(smtphost)
    smtp.sendmail(fromaddr, recips, msg.as_string())
    smtp.quit()


def send_system_email(config, deasserted, asserted):
    try:
        emails = config.get('email', 'system').split(',')
    except (configparser.NoOptionError,
            configparser.NoSectionError):
        return

    if not emails:
        return

    body = ('Security System alert.\n' +
            '\n' +
            'The following new flags have been asserted:\n' +
            ('%s\n' % ','.join(asserted)) +
            '\n' +
            'The following flags are now de-asserted:\n' +
            ('%s\n' % ','.join(deasserted)))

    try:
        _send_system_email(config, 'Security System Alert',
                           emails, body)
    except MissingEmailConfig:
        pass


def send_partition_email(config, partition, deasserted, asserted):
    try:
        emails = config.get('partition_%i' % partition.number,
                            'flags').split(',')
    except (configparser.NoOptionError,
            configparser.NoSectionError):
        return

    try:
        ignore = set(config.get('partition_%i' % partition.number,
                                'ignore_flags').split(','))
    except configparser.NoOptionError:
        ignore = set([])

    deasserted = deasserted - ignore
    asserted = asserted - ignore
    if not asserted and not deasserted:
        return
    if not emails:
        return

    body = ('Security System partition %i alert.\n' % partition.number +
            '\n' +
            'The following new flags have been asserted:\n' +
            ('%s\n' % ','.join(asserted)) +
            '\n' +
            'The following flags are now de-asserted:\n' +
            ('%s\n' % ','.join(deasserted)))

    try:
        _send_system_email(
            config,
            'Security System Partition %i Alert' % partition.number,
            emails, body)
    except MissingEmailConfig:
        pass


def send_partition_status_email(config, partition, recip_key, sub, message):
    try:
        emails = config.get('partition_%i' % partition.number,
                            recip_key).split(',')
    except (configparser.NoOptionError,
            configparser.NoSectionError):
        return

    if not emails:
        return

    body = 'Security System alert:\n%s' % message
    try:
        _send_system_email(
            config,
            'Security: %s' % sub,
            emails, body)
    except MissingEmailConfig:
        pass


def send_log_event_mail(config, event):
    try:
        alarm_emails = set(config.get('email', 'alarms').split(','))
    except (configparser.NoOptionError, configparser.NoSectionError):
        alarm_emails = set([])

    try:
        alarm_events = set(config.get('email', 'alarm_events').split(','))
    except (configparser.NoOptionError, configparser.NoSectionError):
        alarm_events = set(['Alarm', 'Alarm restore', 'Manual fire',])

    try:
        event_emails = set(config.get('email', 'events').split(','))
    except (configparser.NoOptionError, configparser.NoSectionError):
        event_emails = set([])

    emails = set(event_emails)
    if event.event in alarm_events:
        emails |= alarm_emails

    if not emails:
        return

    body = '%s at %s' % (event.event_string, event.timestamp)

    _send_system_email(
        config, 'Security: %s' % event.event,
        emails, body)



"""
Concord panel general and specific alarm type codes.
"""

ALARM_SPECIFIC_TYPES = {
0: 'Unspecified',
1: 'Fire',
2: 'Fire Panic',
3: 'Police',
4: 'Police Panic',
5: 'Medical',
6: 'Medical Panic',
7: 'Auxiliary',
8: 'Auxiliary Panic',
9: 'Tamper',
10: 'No Activity',
11: 'Suspicion',
12: 'Not used',
13: 'Low Temperature',
14: 'High Temperature',
15: 'Keystroke Violation (Touchpad Tamper)',
16: 'Duress',
17: 'Exit Fault',
18: 'Explosive Gas',
19: 'Carbon Monoxide',
20: 'Environmental',
21: 'Latchkey',
22: 'Equipment Tamper',
23: 'Holdup',
24: 'Sprinkler',
25: 'Heat',
26: 'Siren Tamper',
27: 'Smoke',
28: 'Repeater Tamper',
29: 'Fire Pump Activated',
30: 'Fire Pump Failure',
31: 'Fire Gate Valve',
32: 'Low CO2 Pressure',
33: 'Low Liquid Pressure',
34: 'Low Liquid Level',
35: 'Entry/Exit',
36: 'Perimeter',
37: 'Interior',
38: 'Near (Two Trip, Concord only)',
39: 'Water Alarm',
}

FIRE_SPECIFIC_TYPES = {
0: 'Unspecified',
1: 'Hardwire',
2: 'Ground Fault',
3: 'Device',
4: 'Supervisory',
5: 'Low Battery',
6: 'Tamper',
7: 'SAM',
8: 'Partial Obscurity',
9: 'Jam',
10: 'Zone AC Fail',
11: 'n/u',
12: 'NAC Trouble',
13: 'Analog Zone Trouble',
14: 'Fire Supervisory',
15: 'Pump Fail',
16: 'Fire Gate Valve Closed',
17: 'CO2 Pressure Trouble',
18: 'Liquid Pressure Trouble',
19: 'Liquid Level Trouble',
}

BYPASS_SPECIFIC_TYPES = {
0: 'Direct Bypass',
1: 'Indirect Bypass',
2: 'Swinger Bypass',
3: 'Inhibit',
}

OPENING_SPECIFIC_TYPES = {
0: 'Normal Open',
1: 'Early Open',
2: 'Late Open',
3: 'Fail To Open',
4: 'Open Exception',
5: 'Open Extension',
6: 'Open Using Keyfob/Keyswitch',
7: 'Scheduled Open',
8: 'Remote Open',
}

CLOSING_SPECIFIC_TYPES = {
0: 'Normal Close',
1: 'Early Close',
2: 'Late Close',
3: 'Fail To Close',
4: 'Close Exception',
5: 'Close Extension',
6: 'Close Using Keyfob/Keyswitch',
7: 'Scheduled Close',
8: 'Remote Close',
9: 'Recent Close (Concord only)',
}

PARTITION_CONFIG_CHG_SPECIFIC_TYPES = {
0: 'User Access Code Added',
1: 'User Access Code Deleted',
2: 'User Access Code Changed',
3: 'User Access Code Expired',
4: 'User Code Authority Changed',
5: 'Authority Levels Changed',
6: 'Schedule Changed',
7: 'Arming or O/C Schedule Changed',
8: 'Zone Added',
9: 'Zone Deleted',
}

PARTITION_EVENT_SPECIFIC_TYPES = {
0: 'Schedule On',
1: 'Schedule Off',
2: 'Latchkey On',
3: 'Latchkey Off',
4: 'Smoke Detectors Reset',
5: 'Valid User Access Code Entered',
6: 'Arming Level Changed',
7: 'Alarm Reported',
8: 'Agent Release',
9: 'Agent Release Restoral',
10: 'Partition Remote Access',
11: 'Keystroke Violation in Partition',
12: 'Manual Force Arm',
13: 'Auto Force Arm',
14: 'Auto Force Arm Failed',
15: 'Arming Protest Begun',
16: 'Arming Protest Ended',
}
PARTITION_TEST_SPECIFIC_TYPES = {
0: 'Manual Phone Test',
1: 'Auto Phone Test',
2: 'Auto Phone Test with existing trouble',
3: 'Phone Test OK',
4: 'Phone Test Failed',
5: 'User Sensor Test Started',
6: 'User Sensor Test Ended',
7: 'User Sensor Test Completed',
8: 'User Sensor Test Incomplete',
9: 'user Sensor Test Trip',
10: 'Installer Sensor Test Started',
11: 'Installer Sensor Test Ended',
12: 'Installer Sensor Test Completed',
13: 'Installer Sensor Test Incomplete',
14: 'Installer Sensor Test Trip',
15: 'Fire Drill Started',
}

SYSTEM_TROUBLE_SPECIFIC_TYPES = {
0: 'Bus Receiver Failure',
1: 'Bus Antenna Tamper',
2: 'Main Low Battery',
3: 'SnapCard Low Battery',
4: 'Module Low Battery',
5: 'Main AC Failure',
6: 'SnapCard AC Failure',
7: 'Module AC Failure',
8: 'Aux. Power Failure',
9: 'Bus Shutdown',
10: 'Bus Low Power Mode',
11: 'Phone Line 1 Failure',
12: 'Phone Line 2 Failure',
13: 'Remote Phone Tamper',
14: 'Watchdog Reset',
15: 'RAM Failure',
16: 'Flash Failure',
17: 'Printer Error',
18: 'History Buffer (almost) Full',
19: 'History Buffer Overflow',
20: 'Report Buffer Overflor',
21: 'Bus Device Failure',
22: 'Failure To Communicate',
23: 'Long Range Radio Trouble',
24: 'Module Tamper Trouble',
25: 'Un-enrolled Modulte Trouble',
26: 'Audio Output Trouble',
27: 'Analog Module Trouble',
28: 'Cell Module Trouble',
29: 'Buddy 1 Failure',
30: 'Buddy 2 Failure',
31: 'Buddy 3 Failure',
32: 'Buddy 4 Failure',
33: 'SnapCard Trouble',
34: 'Analog Loop Short',
35: 'Analog Loop Break',
36: 'Analog Address 0',
37: 'Un-enrolled Analog Head',
39: 'Duplicate Analog Head',
40: 'Microphone Switch Trouble',
41: 'Microphone Trouble',
42: 'Microhone Wiring Trouble',
43: 'JTECH Premise Paging Trouble',
44: 'Voice Siren Tamper Trouble',
45: 'Microburst Transmit Failure',
46: 'Microbust Transmit Disable',
47: 'Micorburst Module Failure',
48: 'Microburst Not In Service',
49: 'Automation Supervisory Trouble',
50: 'Microburst Module Initializing',
51: 'Printer Paper Out Trouble',
}

SYSTEM_CONFIG_CHG_SPECIFIC_TYPES = {
0: 'Program Mode Entry',
1: 'Program Mode Exit Without Change',
2: 'Program Mode Exit With Change',
3: 'Downloader Session Start',
4: 'Downloader Session End Without Change',
5: 'Downloader Session End With Change',
6: 'Downloader Error',
7: 'Downloader Connection Denied',
8: 'Date/Time Changed',
9: 'Module Added',
10: 'Module Deleted',
11: 'Speech Tokens Changed',
12: 'Code Changed',
13: 'Panel First Service (cold reset)',
14: 'Panel Back In Service (warm reset)',
15: 'Installer Code Changed',
}

SYSTEM_EVENT_SPECIFIC_TYPES = {
0: 'Callback Requested',
1: 'Output Activity (not used, see 18.5 & 18.6)',
2: 'Buddy Reception',
3: 'Buddy Transmission Request',
4: 'History Buffer Cleared',
5: 'Output On',
6: 'Output Off',
}

ALARM_CODES = {
    # General type -> (General type name, Specific type -> Specific type name))
    1:  ("Alarm", ALARM_SPECIFIC_TYPES),
    2:  ("Alarm Cancel", ALARM_SPECIFIC_TYPES),
    3:  ("Alarm Restoral", ALARM_SPECIFIC_TYPES),
    4:  ("Fire Trouble", FIRE_SPECIFIC_TYPES),
    5:  ("Fire Trouble Restoral", FIRE_SPECIFIC_TYPES),
    6:  ("Non-Fire Trouble", FIRE_SPECIFIC_TYPES),
    7:  ("Non-Fire Trouble Restoral", FIRE_SPECIFIC_TYPES),
    8:  ("Bypass", BYPASS_SPECIFIC_TYPES),
    9:  ("Unbypass", BYPASS_SPECIFIC_TYPES),
    10: ("Opening", OPENING_SPECIFIC_TYPES),
    11: ("Closing", CLOSING_SPECIFIC_TYPES),
    12: ("Partition Configuration Change", PARTITION_CONFIG_CHG_SPECIFIC_TYPES),
    13: ("Partition Event", PARTITION_EVENT_SPECIFIC_TYPES),
    14: ("Partition Test", PARTITION_TEST_SPECIFIC_TYPES),
    15: ("System Trouble", SYSTEM_TROUBLE_SPECIFIC_TYPES),
    16: ("System Trouble Restoral", SYSTEM_TROUBLE_SPECIFIC_TYPES),
    17: ("System Configuration Change", SYSTEM_CONFIG_CHG_SPECIFIC_TYPES),
    18: ("System Event", SYSTEM_EVENT_SPECIFIC_TYPES),
}



import argparse
import logging
import logging.handlers
import os
import threading

from concord232 import api
from concord232 import concord

LOG_FORMAT = '%(asctime)-15s %(module)s %(levelname)s %(message)s'


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', default='config.ini',
                        metavar='FILE',
                        help='Path to config file')
    parser.add_argument('--debug', default=False, action='store_true',
                        help='Enable debug')
    parser.add_argument('--log', default=None,
                        metavar='FILE',
                        help='Path to log file')
    parser.add_argument('--serial', default=None,
                        metavar='PORT',
                        help='Serial port to open for stream')
    parser.add_argument('--listen', default='0.0.0.0',
                        metavar='ADDR',
                        help='Listen address (defaults to 0.0.0.0 (all interfaces))')
    parser.add_argument('--port', default=5007, type=int,
                        help='Listen port (defaults to 5007)')
    args = parser.parse_args()

    LOG = logging.getLogger()
    LOG.setLevel(logging.DEBUG)
    formatter = logging.Formatter(LOG_FORMAT)
    istty = os.isatty(0)

    if args.debug and not istty:
        debug_handler = logging.handlers.RotatingFileHandler(
            'debug.log',
            maxBytes=1024*1024*10,
            backupCount=3)
        debug_handler.setFormatter(formatter)
        debug_handler.setLevel(logging.DEBUG)
        LOG.addHandler(debug_handler)

    if istty:
        verbose_handler = logging.StreamHandler()
        verbose_handler.setFormatter(formatter)
        verbose_handler.setLevel(args.debug and logging.DEBUG or logging.INFO)
        LOG.addHandler(verbose_handler)

    if args.log:
        log_handler = logging.handlers.RotatingFileHandler(
            args.log,
            maxBytes=1024*1024*10,
            backupCount=3)
        log_handler.setFormatter(formatter)
        log_handler.setLevel(logging.DEBUG)
        LOG.addHandler(log_handler)

    LOG.info('Ready')
    logging.getLogger('connectionpool').setLevel(logging.WARNING)

    if args.serial:
        ctrl = concord.AlarmPanelInterface(args.serial, 0.25, LOG)
    else:
        LOG.error('Serial and baudrate are required')
        return

    api.CONTROLLER = ctrl

    t = threading.Thread(target=ctrl.message_loop)
    t.daemon = True
    t.start()
    
    api.app.run(debug=False, host=args.listen, port=args.port, threaded=True)





class BadMessageException(Exception):
    pass

def ascii_hex_to_byte(ascii_bytes):
    """ 
    Returns integer value, which is encoded as hexadecimal in first
    two characters (ascii bytes) of input; assumes *ascii_bytes* is a
    string or array of characters.

    Raises ValueError if there was a problem parsing the hex value.
    """
    assert len(ascii_bytes) >= 2
    return int(ascii_bytes[0] + ascii_bytes[1], 16)
    


def total_secs(td):
    """ *td* is a timedelta object """
    return td.days*3600*24 + td.seconds + td.microseconds/1.0e6





from datetime import datetime
import sys
import time
import traceback


is_py2 = sys.version[0] == '2'
if is_py2:
    import Queue as Queue
else:
    import queue as Queue

import serial


from concord232.concord_commands import RX_COMMANDS, \
    build_cmd_equipment_list, EQPT_LIST_REQ_TYPES, \
    build_dynamic_data_refresh, build_keypress, \
    build_cmd_alarm_trouble, KEYPRESS_CODES

from concord232.concord_helpers import ascii_hex_to_byte, total_secs

CONCORD_MAX_ZONE = 6

CONCORD_BAUD     = 9600
CONCORD_BYTESIZE = serial.EIGHTBITS
CONCORD_STOPBITS = serial.STOPBITS_ONE
CONCORD_PARITY   = serial.PARITY_ODD

CONCORD_MAX_LEN = 58 # includes last-index (length) byte but not checksum

MSG_START = chr(0x0A) # line feed
ACK       = chr(0x06)
NAK       = chr(0x15)

CTRL_CHARS = (ACK, NAK)

# Timeout within which sender expects to receive ACKs, in seconds.
#   inbound = message from us to panel
#   outbound = message from panel to us
ACK_TIMEOUT_INBOUND  = 1.0
ACK_TIMEOUT_OUTBOUND = 2.0 
MAX_RESENDS = 3

STOP = 'STOP'

class CommException(Exception):
    pass

class TimeoutException(CommException):
    pass

class BadEncoding(CommException):
    pass

class BadChecksum(CommException):
    pass

class SerialInterface(object):
    def __init__(self, dev_name, timeout_secs, control_char_cb, logger):
        """ 
        *dev_name* is string name of the device e.g. /dev/cu.usbserial
        *timeout_secs* in fractional seconds; e.g. 0.25 = 250 milliseconds
        """
        self.control_char_cb = control_char_cb
        self.logger = logger
        self.logger.debug("SerialInterface Starting")
        # Ugly debugging hack
        if dev_name == 'fake':
            return
        self.serdev = serial.serial_for_url(dev_name, baudrate=CONCORD_BAUD,
                                    bytesize=CONCORD_BYTESIZE, parity=CONCORD_PARITY,
                                    stopbits=CONCORD_STOPBITS, timeout=timeout_secs,
                                    xonxoff=False, rtscts=False, dsrdtr=False)

    def message_chars_maybe_available(self):
        return self.serdev.inWaiting() > 0

    def wait_for_message_start(self):
        """ 
        Read from the serial port until the message-start character is
        received, discarding other characters.  Special control
        characters are handled with the previously provided handler as
        they are encountered.
        
        Returns MSG_START when that character is read from the port;
        if there is a timeout, returns None.
        """

        byte_read = None
        while byte_read != MSG_START:
            byte_read = self._read1()
            if byte_read == '':
                # Timeout
                return None
            #self.logger.debug("Wait for message start, byte_read=%r" % byte_read)
            if byte_read in CTRL_CHARS:
                self.control_char_cb(byte_read)
            # Discard the unrecognized character
    
        return MSG_START

    def _read1(self):
        c = self.serdev.read(size=1).decode('utf-8')
        return c

    def _try_to_read(self, n):
        """ 
        Try to read *n* message chars from the serial port; if there is a
        timeout raise an exception.  Returns tuple of (message chars, control chars).
        """
        ctrl_chars = [ ]
        chars_read = [ ]
        while len(chars_read) < n:
            one_char = self._read1()
            if one_char == '':
                raise TimeoutException("Timeout in the middle of reading message from the panel")
            if one_char in CTRL_CHARS:
                ctrl_chars.append(one_char)
            else:
                chars_read.append(one_char)
        return chars_read, ctrl_chars
        
    def read_next_message(self):
        """
        Read the next message from the serial port, assuming the
        message-start character has just been read.
        
        Returned message is array of bytes.
        
        It is decoded from the ASCII representation, and includes the
        checksum on the end, and the length byte at the start.  The
        checksum is NOT validated.
        
        A valid message will have at 2 bytes for length & checksum,
        plus at least a single byte for the command code, so 3 or more
        bytes in total.
        
        This function will read as many length bytes as are indicated at
        the start of the message, which may *not* be a valid message, and
        so the message returned from here may be as short as only one byte
        (the length byte).
        
        May raise TimeoutException if there is a timeout while reading the
        message.
        
        If any special control character is encountered while reading the
        message, control_char_cb will be called with that character.
        """
        # Read length; this is is encoded as a hex string with two ascii
        # bytes; the length includes the single checksum byte at the end,
        # which is also encoded as a hex string.
        len_bytes, ctrl_chars = self._try_to_read(2)
        try:
            msg_len = ascii_hex_to_byte(len_bytes)
        except ValueError:
            raise BadEncoding("Invalid length encoding: 0x%x 0x%x" % \
                                  (len_bytes[0], len_bytes[1]))
    
        # Read the rest of the message, including checksum.
        msg_ascii = [' '] * (msg_len + 1) * 2
        msg_ascii[0:2] = len_bytes
        msg_bytes, ctrl_chars2 = self._try_to_read(msg_len * 2)
        msg_ascii[2:] = msg_bytes
        ctrl_chars.extend(ctrl_chars2)
    
        # Handle any control characters; we are assuming it's ok to wait
        # until the end of the message to deal with them, since they can
        # be sent asynchronously with respect to other messages sent by
        # the panel e.g. an ACK to one of our sent messages
        for cc in ctrl_chars:
            self.control_char_cb(cc)
    
        # Decode from ascii hex representation to binary.
        msg_bin = [ 0 ] * (msg_len + 1)
        try:
            for i in range(msg_len + 1):
                msg_bin[i] = ascii_hex_to_byte(msg_ascii[2*i:2*i+2])
        except ValueError:
            raise BadEncoding("Invalid message encoding: %r" % msg_ascii)
    
        return msg_bin

    def write_message(self, msg):
        """ 
        *msg* is a message in binary format, with a valid checksum,
        but no leading message-start character.  This method writes an
        ASCII_encoded message to the port preceded by the
        message-start linefeed character.
        """
        framed_msg = MSG_START + encode_message_to_ascii(msg) 
        #self.logger.debug("write_message: %r" % framed_msg)
        self.serdev.write(framed_msg.encode())
        

    def write(self, data):
        """ Write raw *data* to the serial port. """
        self.serdev.write(data)

    def close(self):
        self.serdev.close()

def compute_checksum(bin_msg):
    """ Compute checksum over all of *bin_msg*. """
    assert len(bin_msg) > 0
    cksum = 0
    for b in bin_msg:
        cksum += b
    return cksum % 256    

def validate_message_checksum(bin_msg):
    """
    *bin_msg* is an array of bytes that have already been decoded from
    the Automation Module ascii format, e.g. an array like [ 0x2A,
    0xF9 ] rather than [ '2', 'A', 'F', '9' ].  *bin_msg* must include
    the checksum on the end and last-index (length) byte at the start,
    but not the message-start linefeed.

    Returns True if checksum is as expected, else False.
    """
    assert len(bin_msg) >= 2
    return compute_checksum(bin_msg[:-1]) == bin_msg[-1]

def update_message_checksum(bin_msg):
    assert len(bin_msg) >= 2
    bin_msg[-1] = compute_checksum(bin_msg[:-1])

def encode_message_to_ascii(bin_msg):
    s = ''
    for b in bin_msg:
        s += '%02x' % b
    return s.upper()

def decode_message_from_ascii(ascii_msg):
    n = len(ascii_msg)
    if n % 2 != 0:
        raise BadEncoding("ASCII message has uneven number of characters.")
    b = [ 0 ] * (n/2)
    for i in range(n/2):
        b[i] = ascii_hex_to_byte(ascii_msg[2*i:2*i+2])
    return b


class AlarmPanelInterface(object):
    def __init__(self, dev_name, timeout_secs, logger):
        self.serial_interface = SerialInterface(dev_name, timeout_secs, \
                                                    self.ctrl_char_cb, logger)
        self.timeout_secs = timeout_secs
        self.logger = logger
        self.logger.debug("Starting")
        self.panel = {}
        self.partitions = {}
        self.zones  = {}
        self.users = {}
        self.master_pin = '0520'
     
        self.display_messages = [];
        # Messages on the transmit queue are in binary format with a
        # valid checksum.
        self.tx_queue = Queue.Queue()

        # This queue hold "fake" synthetic messages that the client
        # can send to itself.  If the panel interface seem messages on
        # this queue, it will 'receive' them.
        self.fake_rx_queue = Queue.Queue()

        self.reset_pending_tx()

        self.message_handlers = { } # Command ID -> list of message handlers for that ID.
        for command_code, (command_id, command_name, parser_fn) \
                in RX_COMMANDS.items():
            self.message_handlers[command_id] = [ ]
        

    def register_message_handler(self, command_id, handler_fn):
        """ 
        *handler_fn* will be passed a dict that is the result of
        parsing the message for the specificed command ID.

        Note: these handlers will be called from in the message loop
        thread, NOT the main thread.
        """
        if command_id not in self.message_handlers:
            raise KeyError("No such command ID %r" % command_id)
        self.message_handlers[command_id].append(handler_fn)

    def ctrl_char_cb(self, cc):
        #self.logger.debug("Ctrl char %r" % cc)
        if cc == ACK:
            if self.tx_pending is None:
                self.logger.debug("Spurious ACK")

            self.reset_pending_tx()
        elif cc == NAK:
            if self.tx_pending is None:
                self.logger.debug("Spurious NAK")
            else:
                self.logger.debug("Possible NAK")
                self.maybe_resend_message("NAK")
        else:
            self.logger.info("Unknown control char 0x%02x" % cc)

    def tx_timeout_exceded(self):
        assert self.tx_pending is not None
        elapsed = datetime.now() - self.tx_time
        return total_secs(elapsed) > ACK_TIMEOUT_INBOUND

    def reset_pending_tx(self):
        self.tx_time = None
        self.tx_pending = None
        self.tx_num_attempts = 0
    
    def send_message(self, msg, retry=False):
        """ 
        Send a message directly to the serial port.  Update pending TX
        state.  If *retry* is True, increment the attempts count,
        otherwise reset it to first attempt.
        """
        self.tx_pending = msg
        if retry:
            self.tx_num_attempts += 1
            self.logger.warn("Resending message, attempt %d: %r" % \
                                  (self.tx_num_attempts, encode_message_to_ascii(msg)))
        else:
            self.tx_num_attempts = 1
            self.logger.debug("Sending message (retry=%d) %r" % \
                     (self.tx_num_attempts, encode_message_to_ascii(msg)))
        self.tx_time = datetime.now()
        self.serial_interface.write_message(msg)

    def maybe_resend_message(self, reason):
        if self.tx_num_attempts >= MAX_RESENDS:
            self.logger.error("Unable to send message (%s), too many attempts (%d): %r" % \
                                  (reason, MAX_RESENDS, encode_message_to_ascii(self.tx_pending)))
            self.reset_pending_tx()
        else:
            self.send_message(self.tx_pending, retry=True)

    # XXX include length bytes in the front?  YES
    def enqueue_msg_for_tx(self, msg):
        """
        Put *msg* on the transmit queue, and append a checksum; *msg*
        is modified.

        This method may be called by the main thread; messages
        enqueued here will be consumed and transmitted by the
        background event-loop thread.
        """
        msg.append(compute_checksum(msg))
        self.tx_queue.put(msg)

    def enqueue_synthetic_msg_for_rx(self, msg):
        """
        Put *msg* on the 'fake' receive queue; it will be 'received'
        by this panel interface object.  The checksum will be
        calculated and appended, but the length byte is required at
        the start of the message. *msg* is modified.
        """
        msg.append(compute_checksum(msg))
        self.fake_rx_queue.put(msg)
        

    def stop_loop(self):
        self.tx_queue.put(STOP)

    def message_loop(self):
        self.logger.debug("Message Loop Starting")        
        #self.request_partitions();
        #time.sleep(1)
        self.request_zones();        
        #time.sleep(1)
        #self.request_users();
        #time.sleep(1)
        self.request_dynamic_data_refresh();
        loop_start_at = datetime.now()
        loop_last_print_at = datetime.now()

        while True:
            # Two parts to loop body: 1) look for and handle any
            # incoming messages, and 2) send out any outgoing
            # messages.

            # Hacky flag variables to avoid spinning fast if there is
            # nothing coming in and nothing going out (the common
            # case...)
            no_inputs = True
            no_outputs = True

            # 
            # Handle any synthetic messages and loop them back to us.
            #
            if not self.fake_rx_queue.empty():
                no_inputs = False
                msg = self.fake_rx_queue.get()
                self.logger.debug("Received synthetic message")
                # Don't need to confirm checksum as we computed it
                # ourselves!
                self.handle_message(msg)

            # 
            # Handle incoming messages.
            #
            # Two part test: the first part will fail right away if
            # there no characters, regardless of the timeout, so we
            # minimize time waiting on messages that won't arrive.
            if self.serial_interface.message_chars_maybe_available() \
                    and self.serial_interface.wait_for_message_start() == MSG_START:
                no_inputs = False

                msg_ok = True
                try:
                    msg = self.serial_interface.read_next_message()
                    #self.logger.debug(msg)
                except CommException as ex:
                    self.send_nak()
                    self.logger.error(repr(ex))
                    continue

                msg_time = datetime.now()

                if len(msg) < 3:
                    # Message too short, need at least length byte,
                    # command byte, and checksum byte.
                    self.send_nak()
                    self.logger.error("Message too short: %r" % encode_message_to_ascii(msg))

                if validate_message_checksum(msg):
                    self.send_ack()
                    self.handle_message(msg)
                else:
                    # Bad checksum
                    self.send_nak()
                    self.logger.error("Bad checksum for message %r" % encode_message_to_ascii(msg))

            # TODO: check here if there is pending input and handle it
            # by looping again, before worrying about sending out any
            # commands.

            #
            # If there is a pending message awaiting ack, see if it needs
            # to be resent.  If there is no pending message (or the
            # pending message timed-out), send what's on the transmit
            # queue.
            #
            if self.tx_pending is not None and self.tx_timeout_exceded():
                no_outputs = False
                self.maybe_resend_message("timeout")
            if self.tx_pending is None and not self.tx_queue.empty():
                no_outputs = False
                msg = self.tx_queue.get()
                if msg == STOP:
                    # Close the serial port once all the pending
                    # messages have been sent.  Because we close it,
                    # we can't rerun message_loop(); we have to create
                    # a new AlarmPanelInterface instance.
                    self.serial_interface.close()
                    return
                self.send_message(msg)

            # If there was nothing to do on this pass through the
            # loop, take a nap...
            if no_inputs and no_outputs:
                time.sleep(self.timeout_secs)

            secs_since_print = total_secs(datetime.now() - loop_last_print_at)
            if secs_since_print > 20:
                self.logger.debug("Looping %d" % \
                                              total_secs(datetime.now() - loop_start_at))
                loop_last_print_at = datetime.now()


    def handle_message(self, msg):
        # Assume we have a good message here.  Command code will
        # either be one or two bytes at offset 1.
        cmd1 = msg[1]
        cmd2 = None
        if len(msg) > 3:
            cmd2 = msg[2]

        # self.log("Handle message %r" % encode_message_to_ascii(msg))

        if cmd1 in RX_COMMANDS:
            command = cmd1
            cmd_str = "0x%02x" % command
        elif (cmd1, cmd2) in RX_COMMANDS:
            command = (cmd1, cmd2)
            cmd_str = "0x%02x/0x%02x" % (command[0], command[1])
        else:
            self.logger.error("Unknown command for message %r" % encode_message_to_ascii(msg))
            return

        command_id, command_name, command_parser = RX_COMMANDS[command]
        if command_parser is None:
            self.logger.debug("No parser for command %s %s" % (command_name, command_id))
            return

        #if command_id in ['SIREN_SYNC','TOUCHPAD']:
        #    return

        if command_id in ('SIREN_SYNC','SIREN_SETUP','SIREN_GO','LIGHTS_STATE'):
            return

        if command_id not in ('TOUCHPAD'):
            self.logger.debug("Handling command %s %s, %s" % \
                                      (cmd_str, command_id, command_parser.__name__))
        

        try:
            decoded_command = command_parser(self,msg)
            if not decoded_command:
                return

            decoded_command['command_id'] = command_id
            
            if 'action' in decoded_command:
                try:
                    #self.logger.info('Try to execute: %r' % decoded_command['action'])
                    func = getattr(self, decoded_command['action'], None)
                    if func is not None:
                        func(decoded_command)
                    else:
                        self.logger.info('Counld not execute: %r' % decoded_command['action'])
                except AttributeError as e:
                    self.logger.info(decoded_command['action'] + ' does not exists')
                    self.logger.info(e)
            
            self.logger.debug(repr(decoded_command))
            for handler in self.message_handlers[command_id]:
                self.logger.debug("Calling handler %r" % handler)
        
        except Exception as ex:
            self.logger.error("Problem handling command %r\n%r" % \
                                  (ex, encode_message_to_ascii(msg)))
            self.logger.error(traceback.format_exc())


    def send_the_master_code(self, msg):
        if self.master_pin is not None:
            keys = []
            for k in self.master_pin:
                self.logger.debug("Adding Key: " + k)
                keys.append(0x00+int(k))
            self.send_keypress(keys)

    def send_nak(self):       
        self.serial_interface.write(NAK.encode())

    def send_ack(self):
        self.serial_interface.write(ACK.encode())

    def request_all_equipment(self):
        msg = build_cmd_equipment_list(request_type=0)
        self.enqueue_msg_for_tx(msg)

    def request_zones(self):
        req = EQPT_LIST_REQ_TYPES['ZONE_DATA']
        msg = build_cmd_equipment_list(request_type=req)
        self.enqueue_msg_for_tx(msg)

    def request_partitions(self):
        req = EQPT_LIST_REQ_TYPES['PART_DATA']
        msg = build_cmd_equipment_list(request_type=req)
        self.enqueue_msg_for_tx(msg)

    def request_users(self):
        req = EQPT_LIST_REQ_TYPES['USER_DATA']
        msg = build_cmd_equipment_list(request_type=req)
        self.enqueue_msg_for_tx(msg)

    def request_dynamic_data_refresh(self):
        msg = build_dynamic_data_refresh()
        self.enqueue_msg_for_tx(msg)

    def send_keypress(self, keys, partition=1, no_check=False):
        msg = build_keypress(keys, partition, area=0, no_check=True)
        self.enqueue_msg_for_tx(msg)
        
    def arm_stay(self,option):
        if option == None:
            self.send_keypress([0x02])
        elif option == 'silent':
            self.send_keypress([0x05, 0x02])
        elif option == 'instant':
            self.send_keypress([0x02, 0x04])

    def arm_away(self,option):
        if option == None:
            self.send_keypress([0x03])
        elif option == 'silent':
            self.send_keypress([0x05, 0x03])
        elif option == 'instant':
            self.send_keypress([0x03, 0x04])

    def send_keys(self, keys, group):
        msg = []
        for k in keys:
            a = list(KEYPRESS_CODES.keys())[list(KEYPRESS_CODES.values()).index(str(k))]
            if group:               
                msg.append(a)
            else:
                self.logger.info("Sending key: %r" % msg)
                self.send_keypress([a])        

        if group:
            self.logger.info("Sending group of keys: %r" % msg)
            self.send_keypress(msg)    
       

    def disarm(self,master_pin):
        self.master_pin = master_pin
        self.send_keypress([0x20])
        
    def inject_alarm_message(self, partition, general_type, specific_type, event_data=0):
        msg = build_cmd_alarm_trouble(partition, "System", 1,
                                      general_type, specific_type)
        self.enqueue_synthetic_msg_for_rx(msg)
        
                
                




import datetime

"""
List of command codes and handler functions for commands sent to and
from the alarm panel, plus code to tect mappings.
"""

from concord232.concord_helpers import BadMessageException, ascii_hex_to_byte
from concord232.concord_tokens import decode_text_tokens
from concord232.concord_alarm_codes import ALARM_CODES

STAR = 0xa
HASH = 0xb

KEYPRESS_CODES = {
    0x00: '0',
    0x01: '1',
    0x02: '2',
    0x03: '3',
    0x04: '4',
    0x05: '5',
    0x06: '6',
    0x07: '7',
    0x08: '8',
    0x09: '9',
    0x0a: '*',
    0x0b: '#',
    0x0c: 'Police Panic',
    0x0d: 'Aux. Panic',
    0x0e: 'Fire Panic',
    0x10: 'Lights On',
    0x11: 'Lights Off',
    0x12: 'Lights Toggle',
    0x13: 'Keyswitch On',
    0x14: 'Keyswitch Off',
    0x15: 'Keyswitch Toggle (not implemented)',
    # 0x16 -> 0x1b are undefined
    0x1c: 'Fire TP - Acknowledge',
    0x1d: 'Fire TP - Silence',
    0x1e: 'Fire TP - Fire Test',
    0x1f: 'Fire TP - Smoke Reset',
    0x20: 'Keyfob Disarm',
    0x21: 'Keyfob Arm',
    0x22: 'Keyfob Lights',
    0x23: 'Keyfob Star',
    0x24: 'Keyfob Arm/Disarm',
    0x25: 'Keyfob Lights/Star',
    0x26: 'Keyfob Long Lights',
    0x27: 'Keyfob Direct Arm to Level 3',
    0x28: 'Keyfob Direct Arm to Level 2',
    0x29: 'Keyfob Arm/Star',
    0x2a: 'Keyfob Disarm/Lights',
    # No 0x2b
    0x2c: 'TP A Key',
    0x30: 'TP B Key',
    0x2d: 'TP C Key',
    0x33: 'TP D Key',
    0x2e: 'TP E Key',
    0x36: 'TP F Key',
}

# Protocol docs say: "Bit 6 = held for a few seconds"; not sure how to
# interpret that here; I think it means all the keyfob codes, plus TP
# C & E keys.

CAPABILITY_CODES = {
    # Code -> (Description, Optional Data Description)
    0x00: ("Power Supervision", None),
    0x01: ("Access Control", None),
    0x02: ("Analog Snmoke", None),
    0x03: ("Audio Listen-In", None),
    0x04: ("SnapCard Supervision", None),
    0x05: ("Microburst", None),
    0x06: ("Dual Phone Line", None),
    0x07: ("Energy Management", None),
    0x08: ("Input Zones", "Number of inputs"),
    0x09: ("Phast/Automation/System Manager", None),
    0x00: ("Phone Interface", None),
    0x0b: ("Relay Outputs", "Number of outputs"),
    0x0c: ("RF Receiver", None),
    0x0d: ("RF Transmitter", None),
    0x0e: ("Parallel Printer", None),
    0x0f: ("Unknown", None),
    0x10: ("LED Touchpad", None),
    0x11: ("1-Line/2-Line/BLT Touchpad", None),
    0x12: ("GUI Touchpad", None),
    0x13: ("Voice Evacuation", None),
    0x14: ("Pager", None),
    0x15: ("Downloadable Code/Data", None),
    0x16: ("JTECH Premise Pager", None),
    0x17: ("Cryptography", None),
    0x18: ("LED Display", None),
}



PANEL_TYPES = {
    0x14: "Concord",
    0x0b: "Concord Express",
    0x1e: "Concord Express 4",
    0x0e: "Concord Euro",
    
    0x0d: "Advent Commercial Fire 250",
    0x0f: "Advent Home Navigator 132",
    0x10: "Advent Commercial Burg 250",
    0x11: "Advent Home Navigator 250",
    0x15: "Advent Commercial Burg 500",
    0x16: "Advent Commercial Fire 500",
    0x17: "Advent Commercial Fire 132",
    0x18: "Advent Commercial Burg 132",
}

PANEL_TYPES_CONCORD = (0x14, 0x0b, 0x1e, 0x0e)

# Concord zone types only
ZONE_TYPES = {
    0: 'Hardwired',
    1: 'RF',
    2: 'RF Touchpad',
}

NORMAL = 'Normal'
TRIPPED = 'Tripped'
FAULTED = 'Faulted'
ALARM   = 'Alarm'
TROUBLE = 'Trouble'
BYPASSED = 'Bypassed'

ZONE_STATES = {
    0: NORMAL,
    1: TRIPPED,
    2: FAULTED,
    4: ALARM,
    8: TROUBLE,
    10: BYPASSED,
}

# Concord user number values only.
USER_NUMBERS = {
    246: "System Master Code",
    247: "Installer Code",
    248: "Dealer Code",
    249: "AVM Code",
    250: "Quick Arm",
    251: "Key Switch Arm",
    252: "System",
}

ARMING_LEVELS = {
    0: 'Zone Test',
    1: 'Off',
    2: 'Home/Perimeter',
    3: 'Away/Full',
    4: 'Night',
    5: 'Silent',
}

# Concord sources
ALARM_SOURCE_TYPE = {
    0: "Bus Device",
    1: "Local Phone",
    2: "Zone",
    3: "System",
    4: "Remote Phone",
}

# Reverse map of alarm source name to type code
ALARM_SOURCE_NAME = dict((v, k) for k, v in ALARM_SOURCE_TYPE.items())

# Concord touchpad message types
TOUCHPAD_MSG_TYPE = {
    0: 'Normal',
    1: 'Broadcast',
    }
    
# Concord arming levels
ARM_LEVEL = {
    1: 'Off',
    2: 'Stay',
    3: 'Away',
    8: 'Phone Test',
    9: 'Sensor Test',
}

FEAT_STATES = {
    0x01: 'Chime',
    0x02: 'Energy saver',
    0x04: 'No delay',
    0x08: 'Latchkey',
    0x10: 'Silent arming',
    0x20: 'Quick arm',
}

EQPT_LIST_REQ_TYPES = {
    'ALL_DATA': 0x00,
    'ZONE_DATA': 0x03,
    'PART_DATA': 0x04,
    'BUS_DEV_DATA': 0x05,
    'BUS_CAP_DATA': 0x06,
    'OUTPUT_DATA': 0x07,
    'USER_DATA': 0x09,
    'SCHED_DATA': 0x0a,
    'EVENT_DATA': 0x0b,
    'LIGHT_ATTACH': 0x0c,
}


def ck_msg_len(msg, cmd, desired_len, exact_len=True):
    """ 
    *desired_len* is the length value that would be in the 'last
    index' byte at the start of the message; actual number of bytes
    will be +1 to account for the length.

    If *exact_len* is True, message must be exactly the desired
    length, otherwise it must be _at least_ the desired length.
    """
    if not exact_len:
        comp = 'at least'
        bad_len = len(msg) < desired_len + 1
    else:
        comp = 'exactly'
        bad_len = len(msg) != desired_len + 1

    if bad_len:
        raise BadMessageException("Message too short for command %r, expected %s %d but got %d" % \
                                      (cmd, comp, desired_len, len(msg)-1))

def bytes_to_num(data):
    """ *data* must be at least 4 bytes long, big-endian order. """
    assert len(data) >= 4
    num = data[3]
    num += ((data[2] << 8)  &      0xff00)
    num += ((data[1] << 16) &    0xff0000)
    num += ((data[0] << 24) &  0xff000000)
    return num
    
def num_to_bytes(num):
    return [ 0xff & (num >> 24), 0xff & (num >> 16), 0xff & (num >> 8), 0xff & num ]

def cmd_panel_type(self,msg):
    ck_msg_len(msg, 0x01, 0x0b)
    assert msg[1] == 0x01, "Unexpected command type 0x02x" % msg[1]
    panel_type = msg[2]
    d = { 'panel_type': PANEL_TYPES.get(panel_type, "Unknown Panel Type 0x%02x" % panel_type) }
    if panel_type in PANEL_TYPES_CONCORD:
        # Interpret Concord hw/sw revision numbers.
        # Really not sure about this. XXX
        d['is_concord'] = True

        if 0 < msg[3] < 27:
            letter = chr(ord('A')-1+msg[3])
        else:
            letter = '?'
        if 0 <= msg[4] <= 9:
            digit = chr(ord('0')+msg[4])
        else:
            digit = '?'
        hw_rev = letter + digit
        sw_rev = (msg[5] << 8) + msg[6]
    else:
        d['is_concord'] = False
        hw_rev = "%d.%d" % (msg[3], msg[4])
        sw_rev = "%d.%d" % (msg[5], msg[6])

    d['hardware_revision'] = hw_rev
    d['software_revision'] = sw_rev
    d['serial_number'] = bytes_to_num(msg[7:])

    self.panel = d;
    return d

def cmd_automation_event_lost(self,msg):
    """ 
    (From protocol docs) Panel's automation buffer has overflowed.
    Automation modules should respond to this with request for Dynamic
    Data Refresh and Full Equipment List Request.
    """
    return { }

def build_state_list(state_code, state_dict):
    states = 'Unknown'
    if state_code in state_dict:
        return state_dict[state_code];
    return states

def cmd_zone_status(self,msg):
    ck_msg_len(msg, 0x21, 0x07)
    assert msg[1] == 0x21, "Unexpected command type 0x02x" % msg[1]
    d = { 'partition_number': msg[2],
          'area_number': msg[3],
          'zone_number': (msg[4] << 8) + msg[5],
          'zone_state': build_state_list(msg[6], ZONE_STATES)
          }
    #Update the status
    identifier = 'p' + str(d['partition_number']) + 'z' + str(d['zone_number'])
    if identifier not in self.zones:
        z = { 'partition_number': d['partition_number'],
          'area_number': d['area_number'],
          'group_number': '',
          'zone_number': d['zone_number'],
          'zone_type': 'Unknown',
          'zone_state': d['zone_state'],
          'zone_text': '',
          'zone_text_tokens': [ ],
          }    
        self.zones[identifier] = z
    else:
        self.zones[identifier]['zone_state'] = d['zone_state']
    return d;

def cmd_zone_data(self, msg):
    ck_msg_len(msg, 0x03, 0x09, exact_len=False)
    assert msg[1] == 0x03, "Unexpected command type 0x02x" % msg[1]
    d = { 'partition_number': msg[2],
          'area_number': msg[3],
          'group_number': msg[4],
          'zone_number': (msg[5] << 8) + msg[6],
          'zone_type': ZONE_TYPES.get(msg[7], 'Unknown'),
          'zone_state': build_state_list(msg[8], ZONE_STATES),
          'zone_text': '',
          'zone_text_tokens': [ ],
          }
    if len(msg) > 0x09 + 1:
        d['zone_text'] = decode_text_tokens(msg[9:-1])
        d['zone_text_tokens'] = msg[9:-1]
    
    identifier = 'p' + str(d['partition_number']) + 'z' + str(d['zone_number'])
    self.zones[identifier] = d
    return d;
    
def cmd_arming_level(self,msg):
    ck_msg_len(msg, (0x22, 0x01), 0x08)
    assert (msg[1], msg[2]) == (0x22, 0x01), "Unexpected command type"
    d = { 'partition_number': msg[3],
          'area_number': msg[4],
          'is_keyfob': msg[5] > 0,
          'user_number_high': msg[5],
          'user_number_low': msg[6],
          }
    un = msg[6]
    if un in USER_NUMBERS:
        user_num = USER_NUMBERS[un]
    elif un <= 229:
        user_num = 'Regular User %d' % un
    elif 230 <= un <= 237:
        user_num  = 'Partition %d Master Code' % (un - 230)
    elif 238 <= un <= 245:
        user_num = 'Partition %d Duress Code' % (un - 238)
    else:
        user_num = 'Unknown Code'
        
    d['user_info'] = user_num
    d['arming_level'] = ARMING_LEVELS.get(msg[7], 'Unknown Arming Level')
    d['arming_level_code'] = msg[7]

    if  d['partition_number'] in self.partitions:
        rd = self.partitions[d['partition_number']] 
        rd['user_info'] = user_num
        rd['arming_level'] = ARMING_LEVELS.get(msg[7], 'Unknown Arming Level')
        rd['arming_level_code'] = msg[7]
        self.partitions[d['partition_number']] = rd

    return d

def decode_alarm_type(gen_code, spec_code):
    if gen_code not in ALARM_CODES:
        return 'Unknown', 'Unknown'
    gen_type, spec_type_dict = ALARM_CODES[gen_code]
    return gen_type, spec_type_dict.get(spec_code, 'Unknown')


def cmd_entry_exit_delay(self,msg):
    assert (msg[1], msg[2]) == (0x22, 0x03), "Unexpected command type"
    ck_msg_len(msg, (0x22, 0x03), 0x08)
    d = { 'partition_number': msg[3],
          'area_number': msg[4],
          'delay_seconds': bytes_to_num([0, 0, msg[6], msg[7]]),
          }
    flags = msg[5]
    bits54 = (flags >> 4) & 0x3 
    bit6 = (flags >> 5) & 1
    bit7 = (flags >> 6) & 1
    v = [ ]
    if bits54 == 0:
        v.append('standard')
    elif bits54 == 1:
        v.append('extended')
    elif bits54 == 2:
        v.append('twice extended')
    if bit6 == 1:
        v.append('exit delay')
    else:
        v.append('entry delay')
    if bit7 == 1:
        v.append('end delay')
    else:
        v.append('start delay')

    d['delay_flags'] = v
    return d;

def cmd_alarm_trouble(self,msg):
    assert (msg[1], msg[2]) == (0x22, 0x02), "Unexpected command type"
    ck_msg_len(msg, (0x22, 0x02), 0x0d)
    d = { 'partition_number': msg[3],
          'area_number': msg[4],
          'source_type': ALARM_SOURCE_TYPE.get(msg[5], 'Unknown Source'),
          'source_number': bytes_to_num([0, msg[6], msg[7], msg[8]]),
          'alarm_general_type_code': msg[9],
          'alarm_specific_type_code': msg[10],
          'event_specific_data': (msg[11] << 8) + msg[12],
    }

    # Get text descriptions
    gen_type, spec_type = decode_alarm_type(msg[9], msg[10])
    d['alarm_general_type'] = gen_type
    d['alarm_specific_type'] = spec_type
    
    return d

def build_cmd_alarm_trouble(partition, source_type, source_number, general_type, specific_type, event_data=0):
    assert source_type in ALARM_SOURCE_NAME
    source_code = ALARM_SOURCE_NAME[source_type]
    msg = [ 0x0d, 0x22, 0x02, partition, 0, source_code ] + \
        num_to_bytes(source_number)[1:] + \
        [ general_type, specific_type ] + \
        num_to_bytes(event_data)[2:]
    assert len(msg) == 0x0d
    return msg

def cmd_touchpad(self,msg):
    assert (msg[1], msg[2]) == (0x22, 0x09), "Unexpected command type"
    ck_msg_len(msg, (0x22, 0x09), 0x06, exact_len=False)
    d = { 'partition_number': msg[3],
          'area_number': msg[4],
          'message_type': TOUCHPAD_MSG_TYPE.get(msg[5], 'Unknown Message Type'),
          'display_text': '',
          'timestamp': datetime.datetime.now(),
          }
    if d['partition_number'] > 1:
        return

    if len(msg) > 0x06:
        d['display_text'] = decode_text_tokens(msg[6:-1])
       
    self.display_messages.append(d)
    return d

def cmd_siren_sync(self,msg):
    return { }

def cmd_partition_data(self,msg):
    assert msg[1] == 0x04, "Unexpected command type"
    ck_msg_len(msg, 0x04, 0x05, exact_len=False)
    d = { 'partition_number': msg[2],
          'area_number': msg[3],
          'arming_level': ARM_LEVEL.get(msg[4], 'Unknown Arming Level'),
          'arming_level_code': msg[4],
          'partition_text': '',
    }
    if len(msg) > 0x05:
        d['partition_text'] = decode_text_tokens(msg[5:-1])

    
    if  d['partition_number'] not in self.zones:
        self.partitions[d['partition_number']] = d        
    else:
        rd = self.partitions[d['partition_number']]
        rd['partition_number'] = msg[2]
        rd['area_number'] =  msg[3]
        rd['arming_level' ] =  ARM_LEVEL.get(msg[4], 'Unknown Arming Level')
        rd['arming_level_code'] = msg[4]
        rd['partition_text'] = ''
        self.partitions[d['partition_number']] = rd 
   
    return d

def bcd_decode(chars):
    val = 0
    for c in chars:
        val = 100*val + 10*((c >> 4) & 0xF) + (c & 0xf)
    return val

def cmd_user_data(self,msg):
    assert msg[1] == 0x09, "Unexpected command type"
    ck_msg_len(msg, 0x09, 0x04, exact_len=False)
    d = { 'user_number': msg[3],
          'user_code': 'Not supplied',
        }
    if len(msg) >= 8:
        d['user_code'] = '%04d' % bcd_decode(msg[5:6])
    return d
    
def cmd_sched_data(self,msg):
    return { }

def cmd_sched_event_data(self,msg):
    return { }

def cmd_light_attach(self,msg):
    return { }

def cmd_siren_setup(self,msg):
    return { }

def cmd_siren_go(self,msg):
    return { }

def cmd_siren_stop(self,msg):
    return { }

def cmd_feat_state(self,msg):
    assert (msg[1], msg[2]) == (0x22, 0x0c), "Unexpected command type"
    ck_msg_len(msg, (0x22, 0x0c), 0x06)
    d = { 'partition_number': msg[3],
          'area_number': msg[4],
          'feature_state': build_state_list(msg[5], FEAT_STATES),
          }
    return d;


def cmd_temp(self,msg):
    return { }

def cmd_time_and_date(self,msg):
    return { }

def cmd_lights_state(self,msg):
    return { }

def cmd_user_lights(self,msg):
    return { }

def cmd_keyfob(self,msg):
    return { }

def cmd_clear_image(self,msg):
    """
    (From protocol docs) This command is sent on panel power up
    initialization and when a communication failure restoral with the
    Automation Module occurs. The Concord will also send this command
    when user or installer programming mode is exited.  This is done
    instead of sending a message for each item as it is changed (user
    code deleted, etc.). The Automation Device should perform an
    Equipment List and Refresh when the Clear Image command is
    received.
    """
    return { }

def cmd_eqpt_list_done(self,msg):
    return { }

def cmd_superbus_dev_data(self,msg):
    return { }

def cmd_superbus_dev_cap(self,msg):
    return { }

def cmd_output_data(self,msg):
    return { }

def build_cmd_equipment_list(request_type=0):
    assert request_type in EQPT_LIST_REQ_TYPES.values()
    if request_type == 0:
        return [ 0x2, 0x2 ]
    else:
        return [ 0x3, 0x2, request_type ]

def build_dynamic_data_refresh():
    return [ 0x02, 0x20 ]

def build_keypress(keys, partition=1, area=0, no_check=False):
    assert len(keys) < 55
    if not no_check:
        for k in keys:
            assert k in KEYPRESS_CODES
    data = [ 4+len(keys), 0x40, partition, area ]
    data.extend(keys)
    return data
    

# These are commands sent by the panel and received by the automation
# device (that is, this software).  Some of these may be send
# autonomously by the panel, and some are sent in response to requests
# from the automation device (e.g. equipment list information).
RX_COMMANDS = {
    # Command code -> (command ID, command display name, command parser)
    0x01: ('PANEL_TYPE', "Panel Type", cmd_panel_type),
    0x02: ('EVENT_LOST', "Automation Event Lost", cmd_automation_event_lost),

    # Following are received in response to equipment list requests.
    0x03: ('ZONE_DATA',    "Zone Data", cmd_zone_data),
    0x04: ('PART_DATA',    "Partition Data", cmd_partition_data),
    0x05: ('BUS_DEV_DATA', "SuperBus Device Data", cmd_superbus_dev_data),
    0x06: ('BUS_CAP_DATA', "SuperBus Device Capabilities Data", cmd_superbus_dev_cap),
    0x07: ('OUTPUT_DATA',  "Output Data", cmd_output_data),

    # 0x08 is sent after all 0x03 and 0x05 (zone & SuperBus device)
    # commands have been sent in response to an equipment list
    # request.
    0x08: ('EQPT_LIST_DONE', "Equipment List Complete", cmd_eqpt_list_done),

    0x09: ('USER_DATA',    "User Data", cmd_user_data),
    0x0a: ('SCHED_DATA',   "Schedule Data", cmd_sched_data),
    0x0b: ('EVENT_DATA',   "Scheduled Event Data", cmd_sched_event_data),
    0x0c: ('LIGHT_ATTACH', "Light to Sensor Attachment", cmd_light_attach),
    # End of equipment list responses.

    0x20: ('CLEAR_IMAGE', "Clear Automation Image", cmd_clear_image),

    0x21: ('ZONE_STATUS',        "Zone Status", cmd_zone_status),
    (0x22, 0x01): ('ARM_LEVEL',  "Arming Level", cmd_arming_level),
    (0x22, 0x02): ('ALARM',      "Alarm/Trouble",  cmd_alarm_trouble),
    (0x22, 0x03): ('DELAY',      "Entry/Exit Delay", cmd_entry_exit_delay),
    (0x22, 0x04): ('SIREN_SETUP', "Siren Setup", cmd_siren_setup),
    (0x22, 0x05): ('SIREN_SYNC',  "Siren Synchronize", cmd_siren_sync),
    (0x22, 0x06): ('SIREN_GO',    "Siren Go", cmd_siren_go),

    (0x22, 0x09): ('TOUCHPAD', "Touchpad Display", cmd_touchpad),

    (0x22, 0x0b): ('SIREN_STOP', "Siren Stop", cmd_siren_stop),

    (0x22, 0x0c): ('FEAT_STATE', "Feature State", cmd_feat_state),
    (0x22, 0x0d): ('TEMP', "Temperature", cmd_temp),
    (0x22, 0x0e): ('TIME', "Time and Date", cmd_time_and_date),

    (0x23, 0x01): ('LIGHTS_STATE', "Lights State Command", cmd_lights_state),
    (0x23, 0x02): ('USER_LIGHTS',  "User Lights Command", cmd_user_lights),
    (0x23, 0x03): ('KEYFOB_CMD',   "Keyfob Command", cmd_keyfob),
}

TX_COMMANDS = {
    # Command code -> (command name, command construction function)
    0x02: ("Full Equipment List Request", build_cmd_equipment_list),
    (0x02, 0x03): ("Single Equipment List Request/Zone Data", build_cmd_equipment_list),
    (0x02, 0x04): ("Single Equipment List Request/Partition Data", build_cmd_equipment_list),
    (0x02, 0x05): ("Single Equipment List Request/Superbus Device Data", build_cmd_equipment_list),
    (0x02, 0x06): ("Single Equipment List Request/Superbus Device Capabilities Data", build_cmd_equipment_list),
    (0x02, 0x07): ("Single Equipment List Request/Output Data", build_cmd_equipment_list),
    # No 0x02 / 0x08
    (0x02, 0x09): ("Single Equipment List Request/User Data", build_cmd_equipment_list),
    (0x02, 0x0a): ("Single Equipment List Request/Schedule Data", build_cmd_equipment_list),
    (0x02, 0x0b): ("Single Equipment List Request/Scheduled Event Data", build_cmd_equipment_list),
    (0x02, 0x0c): ("Single Equipment List Request/Light to Sensor Attachment", build_cmd_equipment_list),
    0x20: ("Dynamic Data Refresh Request", build_dynamic_data_refresh),
    0x40: ("Keypress", build_keypress),
}



import json
import requests
import time


class Client(object):
    def __init__(self, url):
        self._url = url
        self._session = requests.Session()
        self._last_event_index = 0

    def list_zones(self):
        r = self._session.get(self._url + '/zones')
        try:
            return r.json['zones']
        except TypeError:
            return r.json()['zones']

    def list_partitions(self):
        r = self._session.get(self._url + '/partitions')
        try:
            return r.json['partitions']
        except TypeError:
            return r.json()['partitions']

    def arm(self, level, option = None):
        r = self._session.get(
            self._url + '/command',
            params={'cmd': 'arm',
                    'level': level,
                    'option': option})
        return r.status_code == 200

    def disarm(self, master_pin):
        r = self._session.get(
            self._url + '/command',
            params={'cmd': 'disarm',
                    'master_pin': master_pin})
        return r.status_code == 200

    def send_keys(self, keys, group=False):
        r = self._session.get(
            self._url + '/command',
            params={'cmd': 'keys',
                    'keys': keys,
                    'group': group})
        return r.status_code == 200

    def get_version(self):
        r = self._session.get(self._url + '/version')
        if r.status_code == 404:
            return '1.0'
        else:
            return r.json()['version']




TOKENS = {
0x0: '0',
0x1: '1',
0x2: '2',
0x3: '3',
0x4: '4',
0x5: '5',
0x6: '6',
0x7: '7',
0x8: '8',
0x9: '9',
# No 0xa
# No 0xb
0xC: '#',
0xD: ':',
0xE: '/',
0xF: '?',
0x10: '.',
0x11: 'A',
0x12: 'B',
0x13: 'C',
0x14: 'D',
0x15: 'E',
0x16: 'F',
0x17: 'G',
0x18: 'H',
0x19: 'I',
0x1A: 'J',
0x1B: 'K',
0x1C: 'L',
0x1D: 'M',
0x1E: 'N',
0x1F: 'O',
0x20: 'P',
0x21: 'Q',
0x22: 'R',
0x23: 'S',
0x24: 'T',
0x25: 'U',
0x26: 'V',
0x27: 'W',
0x28: 'X',
0x29: 'Y',
0x2A: 'Z',
0x2B: ' ',
0x2C: "'",
0x2D: '-',
0x2E: '_',
0x2F: '*',
0x30: 'AC POWER',
0x31: 'ACCESS',
0x32: 'ACCOUNT',
0x33: 'ALARM',
0x34: 'ALL',
0x35: 'ARM',
0x36: 'ARMING',
0x37: 'AREA',
0x38: 'ATTIC',
0x39: 'AUTO',
0x3A: 'AUXILIARY',
0x3B: 'AWAY',
0x3C: 'BACK',
0x3D: 'BATTERY',
0x3E: 'BEDROOM',
0x3F: 'BEEPS',
0x40: 'BOTTOM',
0x41: 'BREEZEWAY',
0x42: 'BASEMENT',
0x43: 'BATHROOM',
0x44: 'BUS',
0x45: 'BYPASS',
0x46: 'BYPASSED',
0x47: 'CABINET',
0x48: 'CANCELED',
0x49: 'CARPET',
0x4A: 'CHIME',
0x4B: 'CLOSET',
0x4C: 'CLOSING',
0x4D: 'CODE',
0x4E: 'CONTROL',
0x4F: 'CPU',
0x50: 'DEGREES',
0x51: 'DEN',
0x52: 'DESK',
0x53: 'DELAY',
0x54: 'DELETE',
0x55: 'DINING',
0x56: 'DIRECT',
0x57: 'DOOR',
0x58: 'DOWN',
0x59: 'DOWNLOAD',
0x5A: 'DOWNSTAIRS',
0x5B: 'DRAWER',
0x5C: 'DISPLAY',
0x5D: 'DURESS',
0x5E: 'EAST',
0x5F: 'ENERGY SAVER',
0x60: 'ENTER',
0x61: 'ENTRY',
0x62: 'ERROR',
0x63: 'EXIT',
0x64: 'FAIL',
0x65: 'FAILURE',
0x66: 'FAMILY',
0x67: 'FEATURES',
0x68: 'FIRE',
0x69: 'FIRST',
0x6A: 'FLOOR',
0x6B: 'FORCE',
0x6C: 'FORMAT',
0x6D: 'FREEZE',
0x6E: 'FRONT',
0x6F: 'FURNACE',
0x70: 'GARAGE',
0x71: 'GALLERY',
0x72: 'GOODBYE',
0x73: 'GROUP',
0x74: 'HALL',
0x75: 'HEAT',
0x76: 'HELLO',
0x77: 'HELP',
0x78: 'HIGH',
0x79: 'HOURLY',
0x7A: 'HOUSE',
0x7B: 'IMMEDIATE',
0x7C: 'IN SERVICE',
0x7D: 'INTERIOR',
0x7E: 'INTRUSION',
0x7F: 'INVALID',
0x80: 'IS',
0x81: 'KEY',
0x82: 'KITCHEN',
0x83: 'LAUNDRY',
0x84: 'LEARN',
0x85: 'LEFT',
0x86: 'LIBRARY',
0x87: 'LEVEL',
0x88: 'LIGHT',
0x89: 'LIGHTS',
0x8A: 'LVING',
0x8B: 'LOW',
0x8C: 'MAIN',
0x8D: 'MASTER',
0x8E: 'MEDICAL',
0x8F: 'MEMORY',
0x90: 'MIN',
0x91: 'MODE',
0x92: 'MOTION',
0x93: 'NIGHT',
0x94: 'NORTH',
0x95: 'NOT',
0x96: 'NUMBER',
0x97: 'OFF',
0x98: 'OFFICE',
0x99: 'OK',
0x9A: 'ON',
0x9B: 'OPEN',
0x9C: 'OPENING',
0x9D: 'PANIC',
0x9E: 'PARTITION',
0x9F: 'PATIO',
0xA0: 'PHONE',
0xA1: 'POLICE',
0xA2: 'POOL',
0xA3: 'PORCH',
0xA4: 'PRESS',
0xA5: 'QUIET',
0xA6: 'QUICK',
0xA7: 'RECEIVER',
0xA8: 'REAR',
0xA9: 'REPORT',
0xAA: 'REMOTE',
0xAB: 'RESTORE',
0xAC: 'RIGHT',
0xAD: 'ROOM',
0xAE: 'SCHEDULE',
0xAF: 'SCRIPT',
0xB0: 'SEC',
0xB1: 'SECOND',
0xB2: 'SET',
0xB3: 'SENSOR',
0xB4: 'SHOCK',
0xB5: 'SIDE',
0xB6: 'SIREN',
0xB7: 'SLIDING',
0xB8: 'SMOKE',
0xB9: 'Sn',
0xBA: 'SOUND',
0xBB: 'SOUTH',
0xBC: 'SPECIAL',
0xBD: 'STAIRS',
0xBE: 'START',
0xBF: 'STATUS',
0xC0: 'STAY',
0xC1: 'STOP',
0xC2: 'SUPERVISORY',
0xC3: 'SYSTEM',
0xC4: 'TAMPER',
0xC5: 'TEMPERATURE',
0xC6: 'TEMPORARY',
0xC7: 'TEST',
0xC8: 'TIME',
0xC9: 'TIMEOUT',
0xCA: 'TOUCHPAD',
0xCB: 'TRIP',
0xCC: 'TROUBLE',
0xCD: 'UNBYPASS',
0xCE: 'UNIT',
0xCF: 'UP',
0xD0: 'VERIFY',
0xD1: 'VIOLATION',
0xD2: 'WARNING',
0xD3: 'WEST',
0xD4: 'WINDOW',
0xD5: 'MENU',
0xD6: 'RETURN',
0xD7: 'POUND',
0xD8: 'HOME',
0xf9: '\n',
0xfa: '<spc>', # pseudo space
0xfb: '\n',
# Not 0xfc
0xfd: '<bs>', # backspace
0xfe: '<blink>', # blink next token
}

def decode_text_tokens(tokens):
    # Convert token codes to strings, adding appropriate spaces.
    s = ''
    n = len(tokens)
    for i, t in enumerate(tokens):
        # Handle 'backspace' by discarding last character; if previous
        # token was a word token, then we are just removing the
        # trailing space; otherwise we are removing the last letter
        # token.  Not sure if this is exactly the algorithm but it
        # will work for my zone names.
        if t == 0xFD:
            s = s[:-1]
            continue
        c = TOKENS.get(t, '@')
        s += c
        # For 'word' tokens, put a space afterwards, unless it's the last one
        if len(c) > 1 and i < n-1 and c[0] != '<':
            s += ' '
        
    return s



from setuptools import setup

setup(name='concord232',
      version='0.15',
      description='GE Concord 4 RS232 Serial Interface Library and Server',
      author='Jason Carter',
      author_email='jason@jason-carter.net',
      url='http://github.com/JasonCarter80/concord232',
      packages=['concord232'],
      install_requires=['requests', 'stevedore', 'prettytable', 'pyserial', 'flask'],
      scripts=['concord232_server', 'concord232_client']
  )
